From 2a1d9004e1e4eeaab0d88b0f957aff27cbe1e4b2 Mon Sep 17 00:00:00 2001
From: Jean-Paul Saman <jean-paul.saman@m2x.nl>
Date: Wed, 2 May 2012 11:28:32 +0200
Subject: [PATCH 4/7] input core: add status field per elementary stream

VLC has status reporting on the input state, but not on the output state.
This makes it impossible to know if playback actually started completely from
checking stream status states.

For instance a failed audio output cannot be determined programatically. This
patch adds an extra field per ES (in src/input/es_out.c) to determine the state
of that ES (DISABLE, ENABLED, ERROR).

If CreateDecoder() cannot instantiate an output, then it set p_dec->b_error = true.
This is used to determine what the return value for es_out_GetEsState() should be.
If p_dec->b_error is true, then it returns an ES_OUT_STATE_ES_ERROR to the caller.
Else it returns ES_OUT_STATE_ES_ENABLED or ES_OUT_STATE_ES_DISABLED.

A function 'es_out_GetEsState()' is available to query an input for the state of
the primary audio, video and SPU elementary streams.
---
 include/vlc_es_out.h |   18 +++++++++++++
 include/vlc_input.h  |   22 ++++++++++++++++
 src/input/control.c  |   12 +++++++++
 src/input/decoder.c  |    2 +-
 src/input/es_out.c   |   68 +++++++++++++++++++++++++++++++++++++++++++++++--
 5 files changed, 118 insertions(+), 4 deletions(-)

diff --git a/include/vlc_es_out.h b/include/vlc_es_out.h
index cf1abce..3bffe0e 100644
--- a/include/vlc_es_out.h
+++ b/include/vlc_es_out.h
@@ -89,10 +89,23 @@ enum es_out_query_e
     ES_OUT_GET_PCR_SYSTEM, /* arg1=mtime_t *, arg2=mtime_t * res=can fail */
     ES_OUT_MODIFY_PCR_SYSTEM, /* arg1=int is_absolute, arg2=mtime_t, res=can fail */
 
+    /* es_out status 'es_out_status_e' */
+    ES_OUT_GET_ES_STATUS, /* arg1= es_out_id_t* arg2=int (AUDIO_ES/VIDEO_ES/SPU_ES) arg3=es_out_state_e * res=can fail */
+
     /* First value usable for private control */
     ES_OUT_PRIVATE_START = 0x10000,
 };
 
+/**
+ * ES status
+ */
+typedef enum es_out_state_e
+{
+    ES_OUT_STATE_ES_DISABLED = 0,
+    ES_OUT_STATE_ES_ENABLED,
+    ES_OUT_STATE_ES_ERROR,
+} es_out_state_e;
+
 struct es_out_t
 {
     es_out_id_t *(*pf_add)    ( es_out_t *, const es_format_t * );
@@ -156,6 +169,11 @@ static inline int es_out_ControlModifyPcrSystem( es_out_t *out, bool b_absolute,
     return es_out_Control( out, ES_OUT_MODIFY_PCR_SYSTEM, b_absolute, i_system );
 }
 
+static inline int es_out_GetEsState( es_out_t *out, const int i_cat, es_out_state_e *pi_state )
+{
+    return es_out_Control( out, ES_OUT_GET_ES_STATUS, i_cat, pi_state );
+}
+
 /**
  * @}
  */
diff --git a/include/vlc_input.h b/include/vlc_input.h
index 7d8320a..83bd4da 100644
--- a/include/vlc_input.h
+++ b/include/vlc_input.h
@@ -340,6 +340,14 @@ typedef enum input_state_e
     ERROR_S,
 } input_state_e;
 
+/* same order as es_out_state_es_e enum in vlc_es_out.h */
+typedef enum input_es_state_e
+{
+    INPUT_ES_STATE_DISABLED = 0,
+    INPUT_ES_STATE_ENABLED,
+    INPUT_ES_STATE_ERROR,
+} input_es_state_e;
+
 /**
  * Input rate.
  *
@@ -516,6 +524,9 @@ enum input_query_e
     /* External clock managments */
     INPUT_GET_PCR_SYSTEM,   /* arg1=mtime_t *, arg2=mtime_t *       res=can fail */
     INPUT_MODIFY_PCR_SYSTEM,/* arg1=int absolute, arg2=mtime_t      res=can fail */
+
+    /* Ask status of main ES objects */
+    INPUT_GET_STATE_ES,     /* arg1=int (AUDIO/VIDEO/SPU_ES), arg2=es_out state * res=can fail */
 };
 
 /** @}*/
@@ -563,6 +574,17 @@ static inline input_state_e input_GetState( input_thread_t * p_input )
     input_Control( p_input, INPUT_GET_STATE, &state );
     return state;
 }
+
+/**
+ * It will return the state of the current selected elementary streams for this input.
+ */
+static inline input_es_state_e input_GetEsState( input_thread_t *p_input, const int i_cat )
+{
+    input_es_state_e state = INPUT_ES_STATE_ERROR;
+    input_Control( p_input, INPUT_GET_STATE_ES, i_cat, &state );
+    return state;
+}
+
 /**
  * It will add a new subtitle source to the input.
  * Provided for convenience.
diff --git a/src/input/control.c b/src/input/control.c
index d50ea97..b24ec08 100644
--- a/src/input/control.c
+++ b/src/input/control.c
@@ -36,6 +36,8 @@
 #include "resource.h"
 #include "es_out.h"
 
+typedef enum input_es_state_e input_es_state_e;
+typedef enum es_out_state_e es_out_state_e;
 
 static void UpdateBookmarksOption( input_thread_t * );
 
@@ -485,6 +487,16 @@ int input_vaControl( input_thread_t *p_input, int i_query, va_list args )
             return es_out_ControlModifyPcrSystem( p_input->p->p_es_out_display, b_absolute, i_system );
         }
 
+        case INPUT_GET_STATE_ES:
+        {
+            int i_cat = (int)va_arg( args, int );
+            input_es_state_e *pi_state = (input_es_state_e *)va_arg( args, input_es_state_e* );
+            es_out_state_e pi_es_state = (es_out_state_e) 0;
+            int ret = es_out_GetEsState( p_input->p->p_es_out_display, i_cat, &pi_es_state );
+            *pi_state = (input_es_state_e) pi_es_state;
+            return ret;
+        }
+
         default:
             msg_Err( p_input, "unknown query in input_vaControl" );
             return VLC_EGENERIC;
diff --git a/src/input/decoder.c b/src/input/decoder.c
index 0b0cdf2..70861ab 100644
--- a/src/input/decoder.c
+++ b/src/input/decoder.c
@@ -2239,7 +2239,7 @@ static aout_buffer_t *aout_new_buffer( decoder_t *p_dec, int i_samples )
             aout_FormatPrepare( &format );
             if( aout_DecNew( p_aout, &format,
                              &p_dec->fmt_out.audio_replay_gain,
-                             &request_vout ) )
+                             &request_vout ) != 0 )
             {
                 input_resource_RequestAout( p_owner->p_resource, p_aout );
                 p_aout = NULL;
diff --git a/src/input/es_out.c b/src/input/es_out.c
index abcebec..0b6e768 100644
--- a/src/input/es_out.c
+++ b/src/input/es_out.c
@@ -67,6 +67,8 @@ typedef struct
     bool b_selected;
     bool b_scrambled;
 
+    es_out_state_e i_state;
+
     /* Clock for this program */
     input_clock_t *p_clock;
 
@@ -218,6 +220,23 @@ static inline bool EsFmtIsTeletext( const es_format_t *p_fmt )
     return p_fmt->i_cat == SPU_ES && p_fmt->i_codec == VLC_CODEC_TELETEXT;
 }
 
+/* */
+static void EsOutChangeEsState( es_out_t *out, es_out_id_t *es, es_out_state_e i_state )
+{
+    if( es && es->p_pgrm )
+    {
+        const bool b_changed = (es->p_pgrm->i_state != i_state);
+        if( b_changed )
+        {
+            es->p_pgrm->i_state = i_state;
+
+            /* es_out failed report error to input */
+            if( i_state == ES_OUT_STATE_ES_ERROR )
+                input_SendEventState( out->p_sys->p_input, ERROR_S );
+        }
+    }
+}
+
 /*****************************************************************************
  * input_EsOutNew:
  *****************************************************************************/
@@ -247,7 +266,6 @@ es_out_t *input_EsOutNew( input_thread_t *p_input, int i_rate )
     p_sys->b_active = false;
     p_sys->i_mode   = ES_OUT_MODE_NONE;
 
-
     TAB_INIT( p_sys->i_pgrm, p_sys->pgrm );
 
     TAB_INIT( p_sys->i_es, p_sys->es );
@@ -592,8 +610,6 @@ static void EsOutChangePosition( es_out_t *out )
     p_sys->i_preroll_end = -1;
 }
 
-
-
 static void EsOutDecodersStopBuffering( es_out_t *out, bool b_forced )
 {
     es_out_sys_t *p_sys = out->p_sys;
@@ -1033,6 +1049,7 @@ static es_out_pgrm_t *EsOutProgramAdd( es_out_t *out, int i_group )
     /* Init */
     p_pgrm->i_id = i_group;
     p_pgrm->i_es = 0;
+    p_pgrm->i_state = ES_OUT_STATE_ES_DISABLED;
     p_pgrm->b_selected = false;
     p_pgrm->b_scrambled = false;
     p_pgrm->psz_name = NULL;
@@ -1553,6 +1570,7 @@ static void EsCreateDecoder( es_out_t *out, es_out_id_t *p_es )
 {
     es_out_sys_t   *p_sys = out->p_sys;
     input_thread_t *p_input = p_sys->p_input;
+    es_out_state_e i_state = ES_OUT_STATE_ES_ERROR;
 
     p_es->p_dec = input_DecoderNew( p_input, &p_es->fmt, p_es->p_pgrm->p_clock, p_input->p->p_sout );
     if( p_es->p_dec )
@@ -1566,10 +1584,14 @@ static void EsCreateDecoder( es_out_t *out, es_out_id_t *p_es )
             if( p_es->p_dec_record && p_sys->b_buffering )
                 input_DecoderStartBuffering( p_es->p_dec_record );
         }
+        if( !p_es->p_dec->b_error )
+            i_state = ES_OUT_STATE_ES_ENABLED;
     }
 
+    EsOutChangeEsState( out, p_es, i_state );
     EsOutDecoderChangeDelay( out, p_es );
 }
+
 static void EsDestroyDecoder( es_out_t *out, es_out_id_t *p_es )
 {
     VLC_UNUSED(out);
@@ -1577,6 +1599,8 @@ static void EsDestroyDecoder( es_out_t *out, es_out_id_t *p_es )
     if( !p_es->p_dec )
         return;
 
+    EsOutChangeEsState( out, p_es, ES_OUT_STATE_ES_DISABLED );
+
     input_DecoderDelete( p_es->p_dec );
     p_es->p_dec = NULL;
 
@@ -2668,11 +2692,49 @@ static int EsOutControlLocked( es_out_t *out, int i_query, va_list args )
         return VLC_SUCCESS;
     }
 
+    /* es_out status */
+    case ES_OUT_GET_ES_STATUS:
+    {
+        const int i_cat = (int)va_arg( args, int );
+        es_out_state_e *pi_state = (es_out_state_e *)va_arg( args, es_out_state_e * );
+
+        *pi_state = ES_OUT_STATE_ES_DISABLED;
+
+        if( !out && !out->p_sys )
+            return VLC_EGENERIC;
+
+        es_out_id_t *p_es = NULL;
+        switch( i_cat )
+        {
+            case VIDEO_ES:
+                p_es = out->p_sys->p_es_video;
+                break;
+            case AUDIO_ES:
+                p_es = out->p_sys->p_es_audio;
+                break;
+            case SPU_ES:
+                p_es = out->p_sys->p_es_sub;
+                break;
+            default:
+                 return VLC_EGENERIC;
+        }
+        if( !p_es )
+            return VLC_EGENERIC;
+        if( !p_es->p_pgrm )
+            return VLC_EGENERIC;
+
+        *pi_state = !p_es->p_dec->b_error ?
+            p_es->p_pgrm->i_state : ES_OUT_STATE_ES_ERROR;
+
+        return VLC_SUCCESS;
+    }
+
     default:
         msg_Err( p_sys->p_input, "unknown query in es_out_Control" );
         return VLC_EGENERIC;
     }
 }
+
 static int EsOutControl( es_out_t *out, int i_query, va_list args )
 {
     es_out_sys_t *p_sys = out->p_sys;
-- 
1.7.7.6

